"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import os
import stat
import logging
import logging.handlers
import sys
import atexit

global_pipegraph = None
is_remote = False
job_uniquifier = {}  # to singletonize jobs on job_id
func_hashes = {}  # to calculate invarionts on functions in a slightly more efficent manner
reactor_was_started = False

#import gc
#gc.set_debug(gc.DEBUG_UNCOLLECTABLE)

default_logging_handler = logging.handlers.SocketHandler('localhost', 5005)
if os.path.exists('logs'):
    file_logging_handler = logging.FileHandler("logs/ppg_run.txt", mode="w")
else:
    file_logging_handler = None
loggers = {}


def start_logging(module):
    key = 'rem' if is_remote else 'ppg'
    name = "%s.%s" % (key, module)
    if not name in loggers:
        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)
        if default_logging_handler:
            logger.addHandler(default_logging_handler)
        if file_logging_handler is not None:
            logger.addHandler(file_logging_handler)
        loggers[name] = logger
    return loggers[name]


def change_logging_port(port):
    """By default, a running Pipegraph chatters to localhost:5005 via tcp 
    (use utils/log_listener.py to listen).
    If you want it to log to another port, use this function before createing the graph.
    """
    global default_logging_handler
    new_handler = logging.handlers.SocketHandler('127.0.0.1', port)
    for logger in list(loggers.values()):
        logger.removeHandler(default_logging_handler)
        logger.addHandler(new_handler)
    default_logging_handler = new_handler


def flush_logging():
    default_logging_handler.flush()
    if file_logging_handler:
        file_logging_handler.flush()


def output_file_exists(filename):
    """Check if a file exists and its size is > 0"""
    if not os.path.exists(filename):
        return False
    st = os.stat(filename)
    if st[stat.ST_SIZE] == 0:
        return False
    return True


class NothingChanged(Exception):
    """For Invariant communication where
    the invariant value changed, but we don't need to invalidate
    the jobs because of it (and we also want the stored value to be updated).

    This is necessary for the FileChecksumInvariant, the filetime might change,
    then we need to check the checksum. If the checksum matches, we need
    a way to tell the Pipegraph to store the new (filetime, filesize, checksum)
    tuple, without invalidating the jobs.
    """

    def __init__(self, new_value):
        self.new_value = new_value


def assert_uniqueness_of_object(object_with_name_attribute, pipeline=None):
    """Makes certain there is only one object with this class & .name.

    This is necesarry so the pipeline jobs assign their data only to the
    objects you're actually working with."""
    if pipeline is None:
        pipeline = global_pipegraph

    if object_with_name_attribute.name.find('/') != -1:
        raise ValueError("Names must not contain /, it confuses the directory calculations")
    typ = object_with_name_attribute.__class__
    if not typ in pipeline.object_uniquifier:
        pipeline.object_uniquifier[typ] = {}
    if object_with_name_attribute.name in pipeline.object_uniquifier[typ]:
        raise ValueError("Doublicate object: %s, %s" % (typ, object_with_name_attribute.name))
    object_with_name_attribute.unique_id = len(pipeline.object_uniquifier[typ])
    pipeline.object_uniquifier[typ][object_with_name_attribute.name] = True


cpu_count = 0


def CPUs():
    """
    Detects the number of CPUs on a system. Cribbed from pp.
    """
    global cpu_count
    if cpu_count == 0:
        cpu_count = 1  # default
        # Linux, Unix and MacOS:
        if hasattr(os, "sysconf"):
            if "SC_NPROCESSORS_ONLN" in os.sysconf_names:
                # Linux & Unix:
                ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
                if isinstance(ncpus, int) and ncpus > 0:
                    cpu_count = ncpus
            else:  # OSX:
                cpu_count = int(os.popen2("sysctl -n hw.ncpu")[1].read())
         # Windows:
        if "NUMBER_OF_PROCESSORS" in os.environ:
            ncpus = int(os.environ["NUMBER_OF_PROCESSORS"])
            if ncpus > 0:
                cpu_count = ncpus
    return cpu_count


#Compatibility shims from six
"""Copyright (c) 2010-2011 Benjamin Peterson

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""
PY3 = sys.version_info[0] == 3
if PY3:
    def reraise(tp, value, tb=None):
        if tb:
            raise tp.with_traceback(tb)
        raise tp
else:
    def exec_(code, globs=None, locs=None):
        """Execute code in a namespace."""
        if globs is None:
            frame = sys._getframe(1)
            globs = frame.f_globals
            if locs is None:
                locs = frame.f_locals
            del frame
        elif locs is None:
            locs = globs
        exec("""exec code in globs, locs""")

    exec_("""def reraise(tp, value, tb=None):
    raise tp, value, tb
""")
#end shims from six



"""
XRange serialization routines
CTypes Extraction code based on a stack overflow answer by Denis Otkidach  

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""


import ctypes


  
PyObject_HEAD = [
                 ('ob_refcnt', ctypes.c_size_t),
                 ('ob_type', ctypes.c_void_p) ]
class XRangeType(ctypes.Structure):
    _fields_ = PyObject_HEAD + [
        ('start', ctypes.c_long),
        ('step', ctypes.c_long),
        ('len', ctypes.c_long),
    ]

def xrangeToCType(xrangeobj):    
    """Cast a xrange to a C representation of it
    Fields modified in the C Representation effect the xrangeobj
    """    
    return ctypes.cast(ctypes.c_void_p(id(xrangeobj)), ctypes.POINTER(XRangeType)).contents


"""Encoding (generally for placing within a JSON object):"""
    
def encodeMaybeXrange(obj, allowLists = True):
    """Encode an object that might be an xrange
    Supports lists (1 level deep) as well"""
    if isinstance(obj, xrange):
        c_xrange = xrangeToCType(obj)
        return ['xrange', c_xrange.start, c_xrange.step, c_xrange.len]
    if allowLists and isinstance(obj, list):
        return [encodeMaybeXrange(elm, allowLists = False) for elm in obj]
    return obj

def decodeMaybeXrange(obj, allowLists = True):
    """Decode a JSON-encoded object that might be an xrange"""
    if isinstance(obj, list):
        if len(obj) == 4 and obj[0] == 'xrange':  #an xrange object
            outrange = xrange(0)
            c_xrange = xrangeToCType(outrange) #get pointer
            c_xrange.start = obj[1]
            c_xrange.step = obj[2]
            c_xrange.len = obj[3]
            return outrange
        elif allowLists: #decode internal xranges
            return [decodeMaybeXrange(elm, allowLists = False) for elm in obj]
    return obj

def decodeXrangeList(obj):
    """Decode a list of elements and encoded xranges into a single list"""
    
    obj =  decodeMaybeXrange(obj, allowLists = False) #outer layer might be an xrange
    if isinstance(obj, xrange):
        return list(obj) #convert to list
    outlst = []
    for elem in obj:
        if isinstance(elem, list) and len(elem) == 4 and elem[0] == 'xrange':
            outlst.extend(decodeMaybeXrange(elem, allowLists = False))
        else:
            outlst.append(elem)
    return outlst

"""Support for piece-wise xranges - effectively a list of xranges"""
class piecewiseXrange(list):
    def __init__(self):
        self._xrangeMode = False
        
    def toXrangeMode(self):
        """Turn on xrange iterator"""
        self._xrangeMode = True
    
    def myIter(self):
        """Ideally, this would overload __iter__, but that presents massive pickling issues"""
        if self._xrangeMode:
            return self.xrangeIter
        else:
            return self.__iter__ 
    
    def xrangeIter(self):  #use a generator to iterate
        for elm in self.__iter__():
            if isinstance(elm, xrange):
                for subelm in elm:
                    yield subelm
            else:
                yield elm


def maybeXrangeIter(lst):
    """If piecewiseXRange use its custom iterator
    Else revert to default"""
    if isinstance(lst, piecewiseXrange):
        return lst.myIter()()
    else:
        return lst.__iter__()

def filterXrangeList(func, xrange_list):
    """ Input is a list of xranges and integers
        This returns a similarly formatted output where func(elem) evaluates to True
        If an xrange reduces to one element, it just inserts an integer
    """
    
    if not hasattr(xrange_list,'__iter__') or isinstance(xrange_list, xrange):
        xrange_list = [xrange_list]
    
    single_range = 2 #if > 0, then outList is just a single xrange
    no_xrange_output = True #outList has no xranges inserted
    
    outList = piecewiseXrange()
    for elm in xrange_list:
        if isinstance(elm, (int, long)):  #elm is 
            if func(elm):
                outList.append(elm)
                single_range = 0 #individual elements present - so not single xrange
        elif isinstance(elm, xrange):
            cxrange = xrangeToCType(elm)
            step = cxrange.step
            basenum = None
            for num in elm: #iterate through xrange
                if func(num):
                    if basenum == None:
                        basenum = num                        
                else:
                    if basenum != None: #push back xrange
                        if num - step == basenum: #only one element: push an integer
                            outList.append(basenum)
                            single_range = 0
                        else:
                            outList.append(xrange(basenum, num, step))
                            single_range-=1
                            no_xrange_output = False
                        basenum = None
            if basenum != None: #cleanup
                num+=step
                if num - step == basenum: #only one element: push an integer
                    outList.append(basenum)
                    single_range = 0
                else:
                    outList.append(xrange(basenum, num, step))
                    single_range-=1
                    no_xrange_output = False
        else:
            raise TypeError('%s (type %s) is not of type int, long or xrange' % (elm, type(elm)))
        
    if outList:
        if not no_xrange_output:
            if single_range > 0: #only one xrange appended - just return it
                return outList[0]
            else:
                outList.toXrangeMode()
    return outList

def iterateXRangeLimit(obj, limit):
    """Generate xrange lists based on some limit.
    Assumes obj is an xrange, or xrange list, or None
    May also return single xrange objects"""
    def innerGenerator(obj):
        if not obj: #empty list/non case
            yield obj
            return
        
        if isinstance(obj, xrange):
            if len(obj) <= limit:
                yield obj
                return
            #use default algorithm if longer
            obj = [obj]
            
        #main algorithm
        outlist = []
        cnt = 0 #number of items appended
        for xr in obj:
            if isinstance(xr, (int, long)):
                outlist.append(xr)
                cnt+=1
                if cnt >= limit:
                    yield outlist
                    outlist = []
                    cnt =0
            elif isinstance(xr, xrange):
                while True:
                    if len(xr) + cnt <= limit:
                        outlist.append(xr)
                        cnt+=len(xr)
                        break
                    else: #break apart xrange
                        oldlen = len(xr)
                        allowed = limit - cnt
                        c_xrange = xrangeToCType(xr)
                        breakpoint = c_xrange.start+c_xrange.step*allowed
                        to_app = xrange(c_xrange.start, breakpoint, c_xrange.step)
                        outlist.append(to_app)
                        yield outlist
                        outlist = []
                        cnt=0
                        xr = xrange(breakpoint,breakpoint+(oldlen-allowed)*c_xrange.step, c_xrange.step)
                if len(outlist) >= limit:
                    yield outlist
                    outlist = []  
                    
            else:
                raise TypeError('%s (type %s) is not of type int, long or xrange' % (xr, type(xr)))
        if outlist:
            yield outlist
    return innerGenerator(obj)
    
"""
Quick unit test
(TODO: Move to unit tests)
"""

if __name__ == '__main__':
    
    for m in [xrange(0,10,2), xrange(0,1), xrange(0), xrange(0,-2), xrange(0,6), xrange(0,7,3)]:
        me = encodeMaybeXrange(m)
        n = decodeMaybeXrange(me)
        print m, me, n
        
        #split
        print list(iterateXRangeLimit(m,1000))
        print list(iterateXRangeLimit(m,3))
        print [list(x[0]) if type(x) is list else x for x in iterateXRangeLimit(m,3)]
    
    xrl = filterXrangeList(lambda x: x%10, xrange(10,100))
    print xrl
    
    lmt = list(iterateXRangeLimit(xrl,15))
    print lmt
    
    lmt_read = [[list(x) for x in y] for y in lmt]
    lmt2 = [reduce(lambda x, y: x+y, x) for x in lmt_read] 
    print lmt2
    
    lenz = [len(x) for x in lmt2]
    print lenz
    
    print list(iterateXRangeLimit(range(25), 13))
        


from __future__ import print_function

"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import time
from . import util
logger = util.start_logging('RC')
import os
import traceback
import multiprocessing
try: 
    import Queue
    queue = Queue
except ImportError:
    import queue
import sys
try:
    import cPickle
    pickle = cPickle
except ImportError:
    import pickle
from . import ppg_exceptions
import tempfile

try:
    from twisted.internet import reactor
    from twisted.internet.protocol import ClientCreator, ProcessProtocol
    from twisted.protocols import amp
    from . import messages
    twisted_available = True
except ImportError:
        twisted_available = False



class DummyResourceCoordinator:
    """For the calculating slaves. so it throws exceptions..."""


def get_memory_available():
    op = open("/proc/meminfo", 'r')
    d = op.read()
    op.close()
    mem_total = d[d.find('MemTotal:') + len('MemTotal:'):]
    mem_total = mem_total[:mem_total.find("kB")].strip()
    swap_total = d[d.find('SwapTotal:') + len('SwapTotal:'):]
    swap_total = swap_total[:swap_total.find("kB")].strip()
    physical_memory = int(mem_total) * 1024
    swap_memory = int(swap_total) * 1024
    return physical_memory, swap_memory


class LocalSystem:
    """A ResourceCoordinator that uses the current machine,
    up to max_cores_to_use cores of it

    It uses multiprocessing and the LocalSlave
    """

    def __init__(self, max_cores_to_use=util.CPUs()):
        self.max_cores_to_use = max_cores_to_use  # todo: update to local cpu count...
        self.slave = LocalSlave(self)
        self.cores_available = max_cores_to_use
        self.physical_memory, self.swap_memory = get_memory_available()
        self.timeout = 5

    def spawn_slaves(self):
        return {
                'LocalSlave': self.slave
                }

    def get_resources(self):
        res = {
                'LocalSlave': {  # this is always the maximum available - the graph is handling the bookeeping of running jobs
                    'cores': self.cores_available,
                    'physical_memory': self.physical_memory,
                    'swap_memory': self.swap_memory,
                    }
                }
        logger.info('get_resources, result %s - %s' % (id(res), res))
        return res

    def enter_loop(self):
        self.spawn_slaves()
        self.que = multiprocessing.Queue()
        logger.info("Starting first batch of jobs")
        self.pipegraph.start_jobs()
        while True:
            self.slave.check_for_dead_jobs()  # whether time out or or job was done, let's check this...
            self.see_if_output_is_requested()
            try:
                logger.info("Listening to que")
                slave_id, was_ok, job_id_done, stdout, stderr, exception, trace, new_jobs = self.que.get(block=True, timeout=self.timeout)  # was there a job done?
                logger.info("Job returned: %s, was_ok: %s" % (job_id_done, was_ok))
                logger.info("Remaining in que (approx): %i" % self.que.qsize())
                job = self.pipegraph.jobs[job_id_done]
                job.was_done_on.add(slave_id)
                job.stdout = stdout
                job.stderr = stderr
                job.exception = exception
                job.trace = trace
                job.failed = not was_ok
                job.stop_time = time.time()
                if job.failed:
                    try:
                        if job.exception.startswith('STR'.encode('UTF-8')):
                            job.exception = job.exception[3:]
                            raise pickle.UnpicklingError("String Transmission")  # what an ugly control flow...
                        logger.info("Before depickle %s" % type(exception))
                        job.exception = pickle.loads(exception)
                        logger.info("After depickle %s" % type(job.exception))
                        logger.info("exception stored at %s" % (job))
                    except pickle.UnpicklingError:  # some exceptions can't be pickled, so we send a string instead
                        pass
                    if job.exception:
                        logger.info("Exception: %s" % repr(exception))
                        logger.info("Trace: %s" % trace)
                    logger.info("stdout: %s" % stdout)
                    logger.info("stderr: %s" % stderr)
                if not new_jobs is False:
                    if not job.modifies_jobgraph():
                        job.exception = ppg_exceptions.JobContractError("%s created jobs, but was not a job with modifies_jobgraph() returning True" % job)
                        job.failed = True
                    else:
                        new_jobs = pickle.loads(new_jobs)
                        logger.info("We retrieved %i new jobs from %s" % (len(new_jobs), job))
                        self.pipegraph.new_jobs_generated_during_runtime(new_jobs)

                more_jobs = self.pipegraph.job_executed(job)
                #if job.cores_needed == -1:
                    #self.cores_available = self.max_cores_to_use
                #else:
                    #self.cores_available += job.cores_needed
                if not more_jobs:  # this means that all jobs are done and there are no longer any more running...
                    break
                self.pipegraph.start_jobs()

            except (queue.Empty, IOError):  # either timeout, or the que failed
                #logger.info("Timout")
                #for job in self.pipegraph.running_jobs:
                    #logger.info('running %s' % (job, ))
                pass
        self.que.close()
        self.que.join_thread()  # wait for the que to close
        logger.info("Leaving loop")

    def see_if_output_is_requested(self):
        import select
        try:
            if select.select([sys.stdin], [], [], 0)[0]:
                ch = sys.stdin.read(1)  # enter pressed...
                self.pipegraph.print_running_jobs()
                pass
        finally:
            pass
            #termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)


class LocalSlave:

    def __init__(self, rc):
        self.rc = rc
        self.slave_id = 'LocalSlave'
        logger.info("LocalSlave pid: %i (runs in MCP!)" % os.getpid())
        self.process_to_job = {}

    def spawn(self, job):
        logger.info("Slave: Spawning %s" % job.job_id)
        job.start_time = time.time()
        #logger.info("Slave: preqs are %s" % [preq.job_id for preq in job.prerequisites])
        preq_failed = False
        if not job.is_final_job:  # final jobs don't load their (fake) prereqs.
            for preq in job.prerequisites:
                if preq.is_loadable():
                    logger.info("Slave: Loading %s" % preq)
                    if not self.load_job(preq):
                        preq_failed = True
                        break
        if preq_failed:
            self.rc.que.put(
                    (
                        self.slave_id,
                        False,  # failed?
                        job.job_id,  # id...
                        '',  # output
                        '',  # output
                        'Prerequsite failed'.encode('UTF-8'),
                        '',
                        False,
                    ))
        else:
            if job.modifies_jobgraph():
                logger.info("Slave: Running %s in slave" % job)
                self.run_a_job(job)
                logger.info("Slave: returned from %s in slave, data was put" % job)
            else:
                logger.info("Slave: Forking for %s" % job.job_id)
                p = multiprocessing.Process(target=self.run_a_job, args=[job, False])
                job.run_info = "pid = %s" % (p.pid, )
                p.start()
                self.process_to_job[p] = job
                logger.info("Slave, returning to start_jobs")

    def load_job(self, job):  # this executes a load job returns false if an error occured
        stdout = tempfile.SpooledTemporaryFile(mode='w')
        stderr = tempfile.SpooledTemporaryFile(mode='w')
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = stdout
        sys.stderr = stderr
        trace = ''
        new_jobs = False
        try:
            job.load()
            was_ok = True
            exception = None
        except Exception as e:
            trace = traceback.format_exc()
            was_ok = False
            exception = e
            try:
                exception = pickle.dumps(exception)
            except Exception as e:  # some exceptions can't be pickled, so we send a string instead
                exception = str(exception)
        stdout.seek(0, os.SEEK_SET)
        stdout_text = stdout.read()
        stdout.close()
        stderr.seek(0, os.SEEK_SET)
        stderr_text = stderr.read()
        stderr.close()
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        if not was_ok:
            self.rc.que.put(
                    (
                        self.slave_id,
                        was_ok,  # failed?
                        job.job_id,  # id...
                        stdout_text,  # output
                        stderr_text,  # output
                        exception,
                        trace,
                        new_jobs,
                    ))
        return was_ok

    def run_a_job(self, job, is_local=True):  # this runs in the spawned processes, except for job.modifies_jobgraph()==True jobs
        stdout = tempfile.SpooledTemporaryFile(mode='w+')
        stderr = tempfile.SpooledTemporaryFile(mode='w+')
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = stdout
        sys.stderr = stderr
        trace = ''
        new_jobs = False
        util.global_pipegraph.new_jobs = None  # ignore jobs created here.
        try:
            temp = job.run()
            was_ok = True
            exception = None
            if job.modifies_jobgraph():
                new_jobs = self.prepare_jobs_for_transfer(temp)
            elif temp:
                raise ppg_exceptions.JobContractError("Job returned a value (which should be new jobs generated here) without having modifies_jobgraph() returning True")
        except Exception as e:
            trace = traceback.format_exc()
            was_ok = False
            exception = e
            try:
                exception = pickle.dumps(e)
            except Exception as e:  # some exceptions can't be pickled, so we send a string instead
                exception = bytes("STR", 'UTF-8') + bytes(e)
        stdout.seek(0, os.SEEK_SET)
        stdout_text = stdout.read()
        stdout.close()
        stderr.seek(0, os.SEEK_SET)
        stderr_text = stderr.read()
        stderr.close()
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        #logger.info("Now putting job data into que: %s - %s" % (job, os.getpid()))
        self.rc.que.put(
                (
                    self.slave_id,
                    was_ok,  # failed?
                    job.job_id,  # id...
                    stdout_text,  # output
                    stderr_text,  # output
                    exception,
                    trace,
                    new_jobs,
                ))
        if not is_local:
            self.rc.que.close()
            self.rc.que.join_thread()

    def prepare_jobs_for_transfer(self, job_dict):
        """When traveling back, jobs-dependencies are wrapped as strings - this should
        prevent nasty suprises"""
         # package as strings
        for job in job_dict.values():
            job.prerequisites = [preq.job_id for preq in job.prerequisites]
            job.dependants = [dep.job_id for dep in job.dependants]
         # unpackanging is don in new_jobs_generated_during_runtime
        self.rc.pipegraph.new_jobs_generated_during_runtime(job_dict)
        return pickle.dumps({})  # The LocalSlave does not need to serialize back the jobs, it already is running in the space of the MCP

    def check_for_dead_jobs(self):
        remove = []
        for proc in self.process_to_job:
            if not proc.is_alive():
                logger.info("Process ended %s" % proc)
                remove.append(proc)
                if proc.exitcode != 0:  # 0 means everything ok, we should have an answer via the que from the job itself...
                    job = self.process_to_job[proc]
                    self.rc.que.put((
                            self.slave_id,
                            False,
                            job.job_id,
                            'no stdout available',
                            'no stderr available',
                            pickle.dumps(ppg_exceptions.JobDiedException(proc.exitcode)),
                            '',
                            False  # no new jobs
                            ))
        for proc in remove:
            del self.process_to_job[proc]


if twisted_available:
    class LocalTwisted:
        """A ResourceCoordinator that uses the current machine,
        up to max_cores_to_use cores of it

        It uses Twisted and one LocalTwistedSlave
        """

        def __init__(self, max_cores_to_use=8):
            self.max_cores_to_use = max_cores_to_use  # todo: update to local cpu count...
            self.slaves = {}
            self.cores_available = max_cores_to_use
            self.memory_available = 50 * 1024 * 1024 * 1024  # 50 gigs ;), todo, update to actual memory + swap... (see LocalSystem)
            self.timeout = 15

        def spawn_slaves(self):
            if self.slaves:
                raise ValueError("spawn_slaves called twice")
            self.slaves = {
                        'LocalSlave': LocalTwistedSlave(self)
                        }
            return self.slaves

        def get_resources(self):
            return {
                    'LocalSlave': {'cores': self.cores_available,
                        'memory': self.memory_available}
                    }

        def enter_twisted(self, pipe):
             # this happens in a spawned process...
            reactor.run()
            print(('exception in spawned', repr(self.pipegraph.jobs['out/a'].exception)))
            from . import cloudpickle
            pipe.send((cloudpickle.dumps(self.pipegraph.jobs), pickle.dumps(self.pipegraph.invariant_status)))

        def enter_loop(self):
            self.slaves_ready_count = 0
            logger.info("starting reactor")
            if not util.reactor_was_started:
                util.reactor_was_started = True
            parent_conn, child_conn = multiprocessing.Pipe()
            p = multiprocessing.Process(None, self.enter_twisted, args=(child_conn, ))
            p.start()
            jobs_str, invariant_str = parent_conn.recv()
            p.join()
            pickle.loads(jobs_str)  # which should automatically fill in the existing jobs...
            self.pipegraph.invariant_status = pickle.loads(invariant_str)
            #p.start()
            #p.join()
            #if p.exitcode != 0:
                #raise ppg_exceptions.RuntimeError()
            #reactor.run()
             # now, the problem here is that you simply can't restart twisted reactors
             # in ordinary coding, this is hardly a problem, though it is a hazzle to being
             # able to start another pipeline.
             # it does get fairly troublesome in testing.
             # the idea of course would be to fork before calling reactor.run,
             # and not returning until the forked child returns.
             # problem with this is of course how to propegate the exceptions
             # let' multiprocessing do the worring...

        def start_when_ready(self, response, slave_id):  # this get's called when a slave has connected and transmitted the pipegraph...
            status = response['ok']
            logger.info("start_when_ready for %s, status was %s" % (slave_id, status))
            if status:
                self.slaves_ready_count += 1
                if self.slaves_ready_count == len(self.slaves):
                    logger.info("Now calling start_jobs")
                    self.pipegraph.start_jobs()
            else:
                reactor.stop()
                raise ppg_exceptions.CommunicationFailure("Slave %s could not load jobgraph. Exception: %s" % (slave_id, response['exception']))

        def end_when_slaves_down(self, slave_id):
            self.slaves_ready_count -= 1
            if self.slaves_ready_count == 0:
                reactor.stop()

        def slave_connection_failed(self, slave_id):
            logger.info("Slave %s connection failed" % slave_id)
            reactor.callLater(0, lambda: reactor.stop())

        def job_ended(self, slave_id, was_ok, job_id_done, stdout, stderr, exception, trace, new_jobs):
            logger.info("Job returned: %s, was_ok: %s, new_jobs %s" % (job_id_done, was_ok, new_jobs))
            job = self.pipegraph.jobs[job_id_done]
            job.was_done_on.add(slave_id)
            job.stdout = stdout
            job.stderr = stderr
            logger.info('stdout %s' % stdout)
            logger.info('stderr %s' % stderr)
            job.exception = exception
            job.trace = trace
            job.failed = not was_ok
            if job.failed:
                try:
                    logger.info("Before depickle %s" % type(exception))
                    job.exception = pickle.loads(exception)
                    logger.info("After depickle %s" % type(job.exception))
                    logger.info("exception stored at %s" % (job))
                except (pickle.UnpicklingError, EOFError):  # some exceptions can't be pickled, so we send a string instead
                    pass
                if job.exception:
                    logger.info("Exception: %s" % repr(exception))
                    logger.info("Trace: %s" % trace)
                logger.info("stdout: %s" % stdout)
                logger.info("stderr: %s" % stderr)
            if new_jobs is not False:
                if not job.modifies_jobgraph():
                    job.exception = ppg_exceptions.JobContractError("%s created jobs, but was not a job with modifies_jobgraph() returning True" % job)
                    job.failed = True
                else:
                    logger.info("now unpickling new jbos")
                    new_jobs = pickle.loads(new_jobs)
                    logger.info("We retrieved %i new jobs from %s" % (len(new_jobs), job))
                    if new_jobs:  # the local system sends back and empty list, because it has called new_jobs_generated_during_runtime itself (without serializing)
                        self.pipegraph.new_jobs_generated_during_runtime(new_jobs)

            more_jobs = self.pipegraph.job_executed(job)
            if job.cores_needed == -1:
                self.cores_available = self.max_cores_to_use
            else:
                self.cores_available += job.cores_needed
            if not more_jobs:  # this means that all jobs are done and there are no longer any more running, so we can return  now...
                for slave_id in self.slaves:
                    self.slaves[slave_id].shut_down().addCallbacks(lambda res: self.end_when_slaves_down(slave_id),
                            lambda failure: self.end_when_slaves_down(slave_id))
            else:
                self.pipegraph.start_jobs()


    class AMP_Return_Protocol(amp.AMP):
        def __init__(self, slave):
            self.slave = slave

        def job_ended(self, arg_tuple_pickle):
            self.slave.job_returned(arg_tuple_pickle)
            return {'ok': True}
        messages.JobEnded.responder(job_ended)

        def __call__(self):
            return self


    class LocalTwistedSlaveProcess(ProcessProtocol):

        def __init__(self, connectionMade_callback, error_callback):
            self.connectionMade_callback = connectionMade_callback
            self.error_callback = error_callback
            self.called_back = False
            self.ended = False

        def connectionMade(self):
            logger.info("LocalTwistedSlaveProcess started in pid %s" % self.transport.pid)
            self.transport.closeStdin()
            #self.slave.connected()

        def outReceived(self, data):
            print(( 'received stdout from slave', data))
            if not self.called_back:
                self.called_back = True
                self.connectionMade_callback()
            print(data)

        def errReceived(self, data):
            print(('received sttderr from slave', data))

        def processExited(self, status):
            logger.info("Slave process ended, exit code %s" % status.value.exitCode)
            exit_code = status.value.exitCode
            if exit_code != 0:
                self.error_callback()
            self.ended = True

        def processEnded(self, status):
            exit_code = status.value.exitCode
            if exit_code != 0:
                raise ppg_exceptions.CommunicationFailure('Could not connect to slave, check logging to see which one')

        def reactor_is_shutting_down(self):
            if not self.ended:
                self.transport.loseConnection()

            return None


    class LocalTwistedSlave:

        def __init__(self, rc):
            self.rc = rc
            self.slave_id = 'LocalTwistedSlave'
            self.start_subprocess()

        def start_subprocess(self):
            self.magic_key = "%s" % time.time()
            cmd = ['python', os.path.abspath(os.path.join(os.path.dirname(__file__), 'util', 'twisted_slave.py')), self.magic_key]

            def do_connect():
                logger.info("Connecting to 500001")
                ClientCreator(reactor, AMP_Return_Protocol(self)).connectTCP('127.0.0.1', 50001).addCallbacks(
                        self.connected,
                        self.rc.slave_connection_failed
                        )

            try:
                protocol = LocalTwistedSlaveProcess(do_connect, lambda: self.rc.slave_connection_failed(self.slave_id))
                self.process = reactor.spawnProcess(protocol, cmd[0],
                        args=cmd, env=os.environ)
            except:
                reactor.stop()
                raise
            reactor.addSystemEventTrigger('before', 'shutdown', protocol.reactor_is_shutting_down)
            #self.process = subprocess.Popen(cmd, cwd=os.getcwd())  # , stdout=subprocess.PIPE)
            #time.sleep(1)
            #port = self.process.stdout.read()
            #logger.info("Connecting to %s" % port)
            pass

        def terminate_process(self, org):
            logger.info("Shutting down process of %s" % self.slave_id)
            self.process.kill()

        def connected(self, proto):
            logger.info("Transport established")
            self.amp = proto
            self.amp.callRemote(messages.MagicKey).addCallbacks(
                    self.magic_key_check,
                    self.magic_key_command_failed)

        def magic_key_command_failed(self, failure):
            logger.info("Magic key command failed: %s" % failure)
            self.amp.callRemote(messages.ShutDown).addCallbacks(
                    self.rc.slave_connection_failed,
                    self.rc.slave_connection_failed,
                    )

        def magic_key_check(self, result):
            logger.info("Received magic key: %s" % result)
            if not 'key' in result:
                self.transmit_pipegraph_failed()
            if result['key'] != self.magic_key:
                logger.info("wrong magic key, sending shutdown")
                self.amp.callRemote(messages.ShutDown).addCallbacks(
                        self.transmit_pipegraph_failed,
                        self.transmit_pipegraph_failed)
                self.transmit_pipegraph_failed()
            else:
                self.transmit_pipegraph(self.rc.pipegraph).addCallbacks(
                    lambda result: self.rc.start_when_ready(result, self.slave_id),
                    self.transmit_pipegraph_failed)

        def transmit_pipegraph_failed(self, failure):
            self.rc.start_when_ready({'ok': False, 'exception': 'Unknown'}, self.slave_id)

        def transmit_pipegraph(self, pipegraph):
            data = cloudpickle.dumps(pipegraph.jobs, 2)  # must be at least version two for the correct new-style class pickling
             # try:  #TODO: remove this unnecessary check
                #cPickle.loads(data)
            #except:
                #print 'error in reloading pipegraph'
                #reactor.stop()
                #raise
            return self.amp.callRemote(messages.TransmitPipegraph, jobs=data)

        def spawn(self, job):
            logger.info("Slave: Spawning %s" % job.job_id)
            self.amp.callRemote(messages.StartJob, job_id=job.job_id).addErrback(
                    self.job_failed_to_start(job.job_id))

        def job_returned(self, encoded_argument):
            logger.info("job_returned")
            args = pickle.loads(encoded_argument)
            self.rc.job_ended(self.slave_id, *args)

        def job_failed_to_start(self, job_id):
            def inner(failure, job_id=job_id):
                self.rc.job_ended(
                        self.slave_id,
                        False,
                        job_id,
                        'Not Available',
                        'Twisted communication error: %s' % failure,
                        '',
                        '',
                        False)
            return inner

        def shut_down(self):
            logger.info("Sending shut_down to %s" % self.slave_id)
            return self.amp.callRemote(messages.ShutDown).addCallbacks(
                    lambda result: self.terminate_process(result),
                    lambda failure: self.terminate_process(failure))



"""
This class is defined to override standard pickle functionality

The goals of it follow:
-Serialize lambdas and nested functions to compiled byte code
-Deal with main module correctly
-Deal with other non-serializable objects

It does not include an unpickler, as standard python unpickling suffices

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""


import ctypes
import os
import pickle
import struct
import sys
import types
from functools import partial
import itertools
from copy_reg import _extension_registry
import new
import dis
import email

#relevant opcodes
STORE_GLOBAL = chr(dis.opname.index('STORE_GLOBAL'))
DELETE_GLOBAL = chr(dis.opname.index('DELETE_GLOBAL'))
LOAD_GLOBAL = chr(dis.opname.index('LOAD_GLOBAL'))
GLOBAL_OPS = [STORE_GLOBAL, DELETE_GLOBAL, LOAD_GLOBAL]

HAVE_ARGUMENT = chr(dis.HAVE_ARGUMENT)
EXTENDED_ARG = chr(dis.EXTENDED_ARG)




try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
def islambda(func):
    return func.func_name == '<lambda>'
import xrange_helper

#debug variables intended for developer use:
printSerialization = False
printMemoization = False

useForcedImports = True #Should I use forced imports for tracking?      



 
class CloudPickler(pickle.Pickler):

    dispatch = pickle.Pickler.dispatch.copy()
    savedForceImports = False
    
    def __init__(self, file, protocol=None, min_size_to_save= 0):        
        pickle.Pickler.__init__(self,file,protocol)
        self.modules = set() #set of modules needed to depickle
    
    #block broken objects
    def save_unsupported(self, obj, pack=None):
        raise pickle.PicklingError("Cannot pickle objects of type %s" % type(obj))    
    dispatch[buffer] = save_unsupported  
    dispatch[types.GeneratorType] = save_unsupported
    
    #python2.6+ supports slice pickling. some py2.5 extensions might as well.  We just test it    
    try:
        slice(0,1).__reduce__()
    except TypeError: #can't pickle -          
        dispatch[slice] = save_unsupported
        
    #email LazyImporters cannot be saved!
    dispatch[email.LazyImporter] = save_unsupported
    
    #itertools objects do not pickle!
    for v in itertools.__dict__.values():
        if type(v) is type:
            dispatch[v] = save_unsupported

    
    def save_dict(self, obj):
        """hack fix 
        If the dict is a global, deal with it in a special way
        """
        #print 'saving', obj
        if obj is __builtins__:
            self.save_reduce(_get_module_builtins, (), obj=obj)
        else:
            pickle.Pickler.save_dict(self, obj)
    dispatch[pickle.DictionaryType] = save_dict            

    
    def save_module(self, obj, pack=struct.pack):
        """
        Save a module as an import
        """
        #print 'try save import', obj.__name__ 
        self.modules.add(obj)
        #self.save_reduce(__import__,(obj.__name__,{},{},['.']), obj=obj)        
        self.save_reduce(subimport,(obj.__name__,), obj=obj)
    dispatch[types.ModuleType] = save_module    #new type       
    
    def save_codeobject(self, obj, pack=struct.pack):
        """
        Save a code object
        """
        #print 'try to save codeobj: ', obj
        args = (
            obj.co_argcount, obj.co_nlocals, obj.co_stacksize, obj.co_flags, obj.co_code,
            obj.co_consts, obj.co_names, obj.co_varnames, obj.co_filename, obj.co_name,
            obj.co_firstlineno, obj.co_lnotab, obj.co_freevars, obj.co_cellvars
        )
        self.save_reduce(types.CodeType, args, obj=obj)
    dispatch[types.CodeType] = save_codeobject    #new type        
        
    def save_function(self, obj, name=None, pack=struct.pack): 
        """ Registered with the dispatch to handle all function types.  
        
        Determines what kind of function obj is (e.g. lambda, defined at 
        interactive prompt, etc) and handles the pickling appropriately.
        """
        write = self.write

        name = obj.__name__
        modname = pickle.whichmodule(obj, name)
        themodule = sys.modules[modname]

        if modname == '__main__':
            themodule = None                            

        if themodule:
            self.modules.add(themodule)

        # if func is lambda, def'ed at prompt, is in main, or is nested, then 
        # we'll pickle the actual function object rather than simply saving a 
        # reference (as is done in default pickler), via save_function_tuple.
        if islambda(obj) or obj.func_code.co_filename == '<stdin>' or themodule == None:
            #Force server to import modules that have been imported in main
            modList = None
            if themodule == None and not self.savedForceImports:                            
                mainmod = sys.modules['__main__']
                if useForcedImports and hasattr(mainmod,'___pyc_forcedImports__'):
                    modList = list(mainmod.___pyc_forcedImports__)
                self.savedForceImports = True                    
            self.save_function_tuple(obj, modList)
            return
        else:   # func is nested
            klass = getattr(themodule, name, None)
            if klass is None or klass is not obj:
                self.save_function_tuple(obj, [themodule])
                return
                
        if obj.__dict__:
            # essentially save_reduce, but workaround needed to avoid recursion
            self.save(_restore_attr)
            write(pickle.MARK + pickle.GLOBAL + modname + '\n' + name + '\n')
            self.memoize(obj)
            self.save(obj.__dict__)
            write(pickle.TUPLE + pickle.REDUCE)
        else:
            write(pickle.GLOBAL + modname + '\n' + name + '\n')
            self.memoize(obj)
    dispatch[types.FunctionType] = save_function
    
    def save_function_tuple(self, func, forced_imports):
        """  Pickles an actual func object.  
        
        A func comprises: code, globals, defaults, closure, and dict.  We 
        extract and save these, injecting reducing functions at certain points 
        to recreate the func object.  Keep in mind that some of these pieces 
        can contain a ref to the func itself.  Thus, a naive save on these 
        pieces could trigger an infinite loop of save's.  To get around that, 
        we first create a skeleton func object using just the code (this is 
        safe, since this won't contain a ref to the func), and memoize it as 
        soon as it's created.  The other stuff can then be filled in later.
        """
        save = self.save
        write = self.write
        
        # save the modules (if any)                
        if forced_imports:
            write(pickle.MARK)
            save(_modules_to_main)
            #print 'forced imports are', forced_imports        
        
            forced_names = map(lambda m: m.__name__, forced_imports)        
            save((forced_names,))
        
            #save((forced_imports,))
            write(pickle.REDUCE)
            write(pickle.POP_MARK)
        
        code, globals, defaults, closure, dict = CloudPickler.extract_func_data(func)        
        
        save(_fill_function)  # skeleton function updater
        write(pickle.MARK)    # beginning of tuple that _fill_function expects 
        
        # create a skeleton function object and memoize it
        save(_make_skel_func)
        save((code, len(closure)))
        write(pickle.REDUCE)
        self.memoize(func)
        
        # save the rest of the func data needed by _fill_function
        save(globals)
        save(defaults)
        save(closure)
        save(dict)
        write(pickle.TUPLE)
        write(pickle.REDUCE)  # applies _fill_function on the tuple
        
    @staticmethod
    def extract_code_globals(co):
        """
        Find all globals names read or written to by codeblock co
        """
        code = co.co_code
        names = co.co_names        
        out_names = set()
        
        n = len(code)
        i = 0
        extended_arg = 0
        while i < n:
            op = code[i]
            
            i = i+1
            if op >= HAVE_ARGUMENT:
                oparg = ord(code[i]) + ord(code[i+1])*256 + extended_arg
                extended_arg = 0
                i = i+2
                if op == EXTENDED_ARG:
                    extended_arg = oparg*65536L                
                if op in GLOBAL_OPS:
                    out_names.add(names[oparg])
        #print 'extracted', out_names, ' from ', names
        return out_names

    @staticmethod
    def extract_func_data(func):
        """
        Turn the function into a tuple of data necessary to recreate it:
            code, globals, defaults, closure, dict
        """
        code = func.func_code
    
        # extract all global ref's
        func_global_refs = CloudPickler.extract_code_globals(code)
        if code.co_consts:   # see if nested function have any global refs
            for const in code.co_consts:
                if type(const) is types.CodeType and const.co_names:
                    func_global_refs = func_global_refs.union( CloudPickler.extract_code_globals(const))
        # process all variables referenced by global environment
        globals = {}
        for var in func_global_refs:
            #Some names, such as class functions are not global - we don't need them
            if func.func_globals.has_key(var):
                globals[var] = func.func_globals[var]                        
        
        # defaults requires no processing
        defaults = func.func_defaults
    
        def get_contents(cell):
            try:
                return cell.cell_contents
            except ValueError, e: #cell is empty error on not yet assigned
                raise pickle.PicklingError('Function to be pickled has free variables that are referenced before assignment in enclosing scope')
                                      
        
        # process closure
        if func.func_closure:
            closure = map(get_contents, func.func_closure)
        else:
            closure = []
            
        # save the dict
        dict = func.func_dict
        
        if printSerialization:
            outvars = ['code: ' + str(code) ]
            outvars.append('globals: ' + str(globals))
            outvars.append('defaults: ' + str(defaults))
            outvars.append('closure: ' + str(closure))        
            print 'function ', func, 'is extracted to: ', ', '.join(outvars)
            
        return (code, globals, defaults, closure, dict)
    
    def save_global(self, obj, name=None, pack=struct.pack):        
        write = self.write
        memo = self.memo

        if name is None:
            name = obj.__name__
        
        modname = getattr(obj, "__module__", None)
        if modname is None:
            modname = pickle.whichmodule(obj, name)

        try:
            __import__(modname)
            themodule = sys.modules[modname]
        except (ImportError, KeyError, AttributeError):  #should never occur                                    
            raise pickle.PicklingError(
                "Can't pickle %r: Module %s cannot be found" %
                (obj, modname))

        if modname == '__main__':
            themodule = None                            

        if themodule:
            self.modules.add(themodule)

        sendRef = True
        typ = type(obj)
        #print 'saving', obj, typ
        try:
            try: #Deal with case when getattribute fails with exceptions
                klass = getattr(themodule, name)
            except (AttributeError):                
                if modname == '__builtin__':  #new.* are misrepeported
                    modname = 'new'
                    __import__(modname)
                    themodule = sys.modules[modname]
                    try:
                        klass = getattr(themodule, name)
                    except AttributeError, a:
                        #print themodule, name, obj, type(obj)
                        raise pickle.PicklingError("Can't pickle builtin %s" % obj)
                else:
                    raise
                            
        except (ImportError, KeyError, AttributeError):                
            if typ == types.TypeType or typ == types.ClassType:
                sendRef = False
            else: #we can't deal with this
                raise
        else:
            if klass is not obj and (typ == types.TypeType or typ == types.ClassType):
                sendRef = False
        if not sendRef:
            #note: Third party types might crash this - add better checks!
            d = dict(obj.__dict__) #copy dict proxy to a dict
            d.pop('__dict__',None)
            d.pop('__weakref__',None)
            self.save_reduce(type(obj),(obj.__name__,obj.__bases__, 
                                   d),obj=obj)
            return

        if self.proto >= 2:
            code = _extension_registry.get((modname, name))
            if code:
                assert code > 0
                if code <= 0xff:
                    write(pickle.EXT1 + chr(code))
                elif code <= 0xffff:
                    write("%c%c%c" % (pickle.EXT2, code&0xff, code>>8))
                else:
                    write(pickle.EXT4 + pack("<i", code))
                return
              
        write(pickle.GLOBAL + modname + '\n' + name + '\n')
        self.memoize(obj)
    dispatch[types.ClassType] = save_global
    dispatch[types.BuiltinFunctionType] = save_global
    dispatch[types.TypeType] = save_global      
    
    def save_instancemethod(self, obj):
        #Memoization rarely is ever useful due to python bounding 
        self.save_reduce(types.MethodType, (obj.im_func, obj.im_self,obj.im_class), obj=obj)
    dispatch[types.MethodType] = save_instancemethod
    
    def save_inst_logic(self, obj):
        """Inner logic to save instance. Based off pickle.save_inst
        Supports __transient__"""
        cls = obj.__class__

        memo  = self.memo
        write = self.write
        save  = self.save

        if hasattr(obj, '__getinitargs__'):
            args = obj.__getinitargs__()
            len(args) # XXX Assert it's a sequence
            pickle._keep_alive(args, memo)
        else:
            args = ()

        write(pickle.MARK)

        if self.bin:
            save(cls)
            for arg in args:
                save(arg)
            write(pickle.OBJ)
        else:
            for arg in args:
                save(arg)
            write(pickle.INST + cls.__module__ + '\n' + cls.__name__ + '\n')

        self.memoize(obj)

        try:
            getstate = obj.__getstate__
        except AttributeError:
            stuff = obj.__dict__            
            #remove items if transient
            if hasattr(obj, '__transient__'):
                transient = obj.__transient__
                stuff = stuff.copy()
                for k in list(stuff.keys()):
                    if k in transient:
                        del stuff[k]            
        else:
            stuff = getstate()
            pickle._keep_alive(stuff, memo)
        save(stuff)
        write(pickle.BUILD)
        
    
    def save_inst(self, obj):
        #Hack to detect PIL Image instances without importing Imaging
        if hasattr(obj,'im') and hasattr(obj,'palette') and 'Image' in obj.__module__:
            self.save_image(obj)
        else:
            self.save_inst_logic(obj)        
    dispatch[types.InstanceType] = save_inst
    
    def save_reduce(self, func, args, state=None,
                    listitems=None, dictitems=None, obj=None):
        """Modified to support __transient__ on new objects
        Change only affects protocol level 2 (which is always used by PiCloud"""        
        # Assert that args is a tuple or None
        if not isinstance(args, types.TupleType):
            raise pickle.PicklingError("args from reduce() should be a tuple")

        # Assert that func is callable
        if not hasattr(func, '__call__'):
            raise pickle.PicklingError("func from reduce should be callable")

        save = self.save
        write = self.write

        # Protocol 2 special case: if func's name is __newobj__, use NEWOBJ
        if self.proto >= 2 and getattr(func, "__name__", "") == "__newobj__":
            #Added fix to allow transient
            cls = args[0]
            if not hasattr(cls, "__new__"):
                raise pickle.PicklingError(
                    "args[0] from __newobj__ args has no __new__")
            if obj is not None and cls is not obj.__class__:
                raise pickle.PicklingError(
                    "args[0] from __newobj__ args has the wrong class")
            args = args[1:]
            save(cls)
            
            #Don't pickle transient entries
            if hasattr(obj, '__transient__'):
                transient = obj.__transient__                
                state = state.copy()
                
                for k in list(state.keys()):
                    if k in transient:
                        del state[k]
            
            save(args)
            write(pickle.NEWOBJ)
        else:
            save(func)
            save(args)
            write(pickle.REDUCE)

        if obj is not None:
            self.memoize(obj)

        # More new special cases (that work with older protocols as
        # well): when __reduce__ returns a tuple with 4 or 5 items,
        # the 4th and 5th item should be iterators that provide list
        # items and dict items (as (key, value) tuples), or None.

        if listitems is not None:
            self._batch_appends(listitems)

        if dictitems is not None:
            self._batch_setitems(dictitems)

        if state is not None:
            save(state)
            write(pickle.BUILD)

    
    def save_xrange(self, obj):
        """Save an xrange object in python 2.5
        Python 2.6 supports this natively
        Code based on a stackoverflow answer from Denis Otkidach"""        
        c_range_obj = xrange_helper.xrangeToCType(obj)
        self.save_reduce(_build_xrange,(c_range_obj.start, c_range_obj.step, c_range_obj.len))
    
    #python2.6+ supports xrange pickling. some py2.5 extensions might as well.  We just test it    
    try:
        xrange(0).__reduce__()
    except TypeError: #can't pickle -- use PiCloud pickler    
        dispatch[xrange] = save_xrange
        
    def save_partial(self, obj):
        """Partial objects do not serialize correctly in python2.x -- this fixes the bugs"""
        self.save_reduce(_genpartial, (obj.func, obj.args, obj.keywords))
    
    dispatch[partial] = save_partial
    
        

    def save_file(self, obj):        
        """Save a file"""        
        import StringIO as pystringIO #we can't use cStringIO as it lacks the name attribute
        from ..transport.adapter import SerializingAdapter
        
        if not hasattr(obj, 'name') or  not hasattr(obj, 'mode'):        
            raise pickle.PicklingError("Cannot pickle files that do not map to an actual file")
        if obj.name == '<stdout>':
            return self.save_reduce(getattr, (sys,'stdout'), obj=obj)
        if obj.name == '<stderr>':
            return self.save_reduce(getattr, (sys,'stderr'), obj=obj)
        if obj.name == '<stdin>':
            raise pickle.PicklingError("Cannot pickle standard input")        
        if  hasattr(obj, 'isatty') and obj.isatty():
            raise pickle.PicklingError("Cannot pickle files that map to tty objects")
        if 'r' not in obj.mode:
            raise pickle.PicklingError("Cannot pickle files that are not opened for reading")        
        name = obj.name
        try:
            fsize = os.stat(name).st_size
        except OSError:
            raise pickle.PicklingError("Cannot pickle file %s as it cannot be stat" % name)    

        if obj.closed:
            #create an empty closed string io
            retval = pystringIO.StringIO("")
            retval.close()
        elif not fsize: #empty file
            retval = pystringIO.StringIO("")                        
            try:
                tmpfile = file(name)
                tst = tmpfile.read(1)
            except IOError:
                raise pickle.PicklingError("Cannot pickle file %s as it cannot be read" % name)
            tmpfile.close()
            if tst != '':
                raise pickle.PicklingError("Cannot pickle file %s as it does not appear to map to a physical, real file" % name)
        elif fsize > SerializingAdapter.maxTransmitData:
            raise pickle.PicklingError("Cannot pickle file %s as it exceeds cloudconf.py's max_transmit_data of %d" % 
                                       (name,SerializingAdapter.maxTransmitData))
        else:            
            try:
                tmpfile = file(name)
                contents = tmpfile.read(SerializingAdapter.maxTransmitData)
                tmpfile.close()
            except IOError:
                raise pickle.PicklingError("Cannot pickle file %s as it cannot be read" % name)
            retval = pystringIO.StringIO(contents)
            curloc = obj.tell()
            retval.seek(curloc)
                
        retval.name = name
        self.save(retval)  #save stringIO
        self.memoize(obj)
                
    dispatch[file] = save_file
    """Special functions for Add-on libraries"""
    
    """numpy ufunc hack"""
    try:
        import numpy
        numpy_tst_mods = ['numpy', 'scipy.special']           
        
        def save_ufunc(self, obj):
            name = obj.__name__
            for tst_mod_name in self.numpy_tst_mods:
                tst_mod = sys.modules.get(tst_mod_name, None)
                if tst_mod:
                    if name in tst_mod.__dict__:
                        self.save_reduce(_getobject, (tst_mod_name, name))
                        return
            raise pickle.PicklingError('cannot save %s. Cannot resolve what module it is defined in' % str(obj))
            
        dispatch[numpy.ufunc] = save_ufunc        
        
    except ImportError:
        pass
    
    """scikits.timeseries hack"""
    try:
        import scikits.timeseries.tseries as ts
        
        def save_timeseries(self, obj):
            import scikits.timeseries.tseries as ts
               
            func, reduce_args, state = obj.__reduce__()
            if func != ts._tsreconstruct:
                raise pickle.PicklingError('timeseries using unexpected reconstruction function %s' % str(func))
            state = (1,
                             obj.shape,
                             obj.dtype,
                             obj.flags.fnc,
                             obj._data.tostring(),
                             ts.getmaskarray(obj).tostring(),
                             obj._fill_value,
                             obj._dates.shape,
                             obj._dates.__array__().tostring(),
                             obj._dates.dtype, #added -- preserve type
                             obj.freq,                             
                             obj._optinfo,
                             )                        
            return self.save_reduce(_genTimeSeries, (reduce_args, state))
                
        dispatch[ts.TimeSeries] = save_timeseries
        
    except ImportError:
        pass
    
    
    """Python Imaging Library"""
    def save_image(self, obj):
        if not obj.im and obj.fp and 'r' in obj.fp.mode and obj.fp.name \
            and not obj.fp.closed and (not hasattr(obj, 'isatty') or not obj.isatty()):
            #if image not loaded yet -- lazy load
            self.save_reduce(_lazyloadImage,(obj.fp,), obj=obj)        
        else:
            #image is loaded - just transmit it over
            self.save_reduce(_generateImage, (obj.size, obj.mode, obj.tostring()), obj=obj)

    """
    def memoize(self, obj):
        pickle.Pickler.memoize(self, obj)
        if printMemoization:
            print 'memoizing ' + str(obj)
    """
    
    

# Shorthands for legacy support

def dump(obj, file, protocol=2):
    CloudPickler(file, protocol).dump(obj)

def dumps(obj, protocol=2):
    file = StringIO()
        
    cp = CloudPickler(file,protocol)
    cp.dump(obj)
    
    #print 'cloud dumped', str(obj), str(cp.modules)
        
    return file.getvalue()


#hack for __import__ not working as desired
def subimport(name):
    __import__(name)
    return sys.modules[name]

# restores function attributes
def _restore_attr(obj, attr):
    for key, val in attr.items():
        setattr(obj, key, val)
    return obj

def _get_module_builtins():
    return pickle.__builtins__

def _modules_to_main(modList):
    """Force every module in modList to be placed into main"""
    if not modList:
        return
    
    main = sys.modules['__main__']
    for modname in modList:
        if type(modname) is str:         
            try:
                mod = __import__(modname)
            except ImportError, i:
                sys.stderr.write('warning: could not import %s\n.  Your function may unexpectedly error due to this import failing; \
A version mismatch is likely.  Specific error was %s\n\n' % (modname, str(i)))
            else:
                setattr(main,mod.__name__, mod)
        else:
            #REVERSE COMPATIBILITY FOR CLOUD CLIENT 1.5 (WITH EPD)
            #In old version actual module was sent            
            setattr(main,modname.__name__, modname)

#object generators:
def _build_xrange(start, step, len):
    """Built xrange explicitly"""
    baserange = xrange(0)    
    c_range_obj = xrange_helper.xrangeToCType(baserange)
    c_range_obj.start = start
    c_range_obj.step = step
    c_range_obj.len = len
    return baserange

def _genpartial(func, args, kwds):
    if not args:
        args = ()
    if not kwds:
        kwds = {}
    return partial(func, *args, **kwds)
 

def _fill_function(func, globals, defaults, closure, dict):
    """ Fills in the rest of function data into the skeleton function object
        that were created via _make_skel_func().
         """
    func.func_globals.update(globals)
    func.func_defaults = defaults
    func.func_dict = dict
    
    if len(closure) != len(func.func_closure):
        raise pickle.UnpicklingError("closure lengths don't match up")    
    for i in range(len(closure)):
        _change_cell_value(func.func_closure[i], closure[i])
            
    return func

def _make_skel_func(code, num_closures):
    """ Creates a skeleton function object that contains just the provided
        code and the correct number of cells in func_closure.  All other
        func attributes (e.g. func_globals) are empty.
    """
    #build closure (cells):
    cellnew = ctypes.pythonapi.PyCell_New
    cellnew.restype = ctypes.py_object
    cellnew.argtypes = (ctypes.py_object,)
    dummy_closure = tuple(map(lambda i: cellnew(None), range(num_closures)))
            
    return types.FunctionType(code, {'__builtins__': __builtins__}, 
                              None, None, dummy_closure)
        
# this piece of opaque code is needed below to modify 'cell' contents
cell_changer_code = new.code(
    1, 1, 2, 0,
    ''.join([
        chr(dis.opmap['LOAD_FAST']), '\x00\x00',
        chr(dis.opmap['DUP_TOP']),
        chr(dis.opmap['STORE_DEREF']), '\x00\x00',
        chr(dis.opmap['RETURN_VALUE'])
    ]), 
    (), (), ('newval',), '<nowhere>', 'cell_changer', 1, '', ('c',), ()
)

def _change_cell_value(cell, newval):
    """ Changes the contents of 'cell' object to newval """
    return new.function(cell_changer_code, {}, None, (), (cell,))(newval)

"""Constructors for 3rd party libraries"""

def _getobject(modname, attribute):
    mod = __import__(modname)
    return mod.__dict__[attribute]

def _generateImage(size, mode, str_rep):
    """Generate image from string representation"""
    import Image
    i = Image.new(mode, size)
    i.fromstring(str_rep)
    return i

def _lazyloadImage(fp):
    import Image
    fp.seek(0)  #works in almost any case
    return Image.open(fp) 

"""Timeseries"""
def _genTimeSeries(reduce_args, state):
    import scikits.timeseries.tseries as ts
    from numpy import ndarray
    from numpy.ma import MaskedArray
    
    
    time_series = ts._tsreconstruct(*reduce_args)
    
    #from setstate modified
    (ver, shp, typ, isf, raw, msk, flv, dsh, dtm, dtyp, frq, infodict) = state
    #print 'regenerating %s' % dtyp
    
    MaskedArray.__setstate__(time_series, (ver, shp, typ, isf, raw, msk, flv))
    _dates = time_series._dates
    #_dates.__setstate__((ver, dsh, typ, isf, dtm, frq))  #use remote typ
    ndarray.__setstate__(_dates,(dsh,dtyp, isf, dtm))
    _dates.freq = frq
    _dates._cachedinfo.update(dict(full=None, hasdups=None, steps=None,
                                   toobj=None, toord=None, tostr=None))
    # Update the _optinfo dictionary
    time_series._optinfo.update(infodict)
    return time_series





"""
The MIT License (MIT)

Copyright (c) 2012, Florian Finkernagel <finkernagel@imt.uni-marburg.de>

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

By default, Machete don't text", er,  "twisted don't fork"
for fear of weirdish signal interactions.

Now, I boldly assume that that can get handled and introduce forking process 
to twisted, which we need to replicate the multiprocessing solution we have
in LocalResourceCoordinator and in the old pypipeline
"""

from twisted.internet import process
import gc
import traceback
import os
import sys

class ForkedProcess(process.Process):

    def __init__(self, reactor, callback, proto):
        process.Process.__init__(self, reactor, callback, (), {}, False,
                proto, uid=None, gid=None, childFDs = None)


    def _execChild(self, path, settingUID, uid, gid, callback, args, environment):
        if args:
            raise ValueError("args not supported")
        if path:
            raise ValueError("Path not supported")
        if settingUID:
            raise ValueError("settingUID not supported")
        #now, this is in the forked child already, and inside the try: except block of _BaseProcess._fork 
        #make sure we get the pipes right...
        sys.stdout = os.fdopen(1, 'w')
        sys.stderr = os.fdopen(2, 'w')
        sys.stdout.flush()
        sys.stderr.flush()
        gc.enable()
        #so... go for it
        try:
            callback()
        except SystemExit as e:
            sys.stdout.flush()
            sys.stderr.flush()
            os._exit(e.code)
        except:
            # If there are errors, bail and try to write something
            # descriptive to the forked process' stderr.
            try:
                #stderr = os.fdopen(2, 'w')
                stderr = sys.stderr
                stderr.write("Upon callback %s" % callback)
                traceback.print_exc(file=stderr)
                sys.stdout.flush()
                sys.stderr.flush()
                for fd in range(3):
                    os.close(fd)
                os._exit(1)
            except:
                pass # make *sure* the child terminates
                sys.stdout.flush()
                sys.stderr.flush()
                os._exit(1)
        sys.stdout.flush()
        sys.stderr.flush()
        os._exit(0)
